/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// CharHdl.h -- character handlers and descendants

#ifndef _charhdl_h_
#define _charhdl_h_

#include "type.h"
#include "stream.h"
#include "list.h"

// character handlers

class CharHdl {
public:
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
    virtual void par();
};

class CmdChar : public CharHdl {
public:
    bool translateCmd(Stream *, char *);
    bool scanCmd(Stream *, char *);
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
    virtual void par();
    virtual SubStream* getCmd(Stream *, char);
};

class IgnoreChar : public CharHdl {
public:
    virtual bool translate(Stream *, char);
    virtual void par();
};

class IllegalChar : public IgnoreChar {
public:
    virtual bool translate(Stream *, char);
};

class EndChar : public CharHdl {
public:
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
    virtual void par();
};

class GermanQuoteChar : public CharHdl {
public:
    virtual bool translate(Stream *, char);
    virtual bool scan(Stream *, char);
};

class UmlautChar : public CharHdl {
public:
    virtual bool translate(Stream *, char);
};

class SubscriptChar : public CharHdl {
public:
  virtual bool translate(Stream*, char);
};

class SuperscriptChar : public CharHdl {
public:
  virtual bool translate(Stream*, char);
};

// character scanner

class Scanner {
protected:
    CharHdl *hdl[256];
    bool parFlag;
    List<int> *endHdl; // <int> is only a hack, should be <void()>
public:
    Scanner();
    void setHdl(char c, CharHdl *h) { hdl[c] = h; }
    CharHdl *getHdl(char c) { return hdl[c]; }
    void translate(Stream *);
    void scan(Stream *);
    void eatWhite(Stream *);
    void atEnd(void (*)());
    virtual void markPar() {}
    virtual void donePar() {}
};

class TextScanner : public Scanner {
public:
    TextScanner();
    virtual void markPar() { parFlag = true; }
    virtual void donePar() { parFlag = false; }
};

class MathScanner : public Scanner {
public:
    MathScanner();
};

class PlainScanner : public Scanner {
public:
    PlainScanner();
};

class MupScanner : public Scanner {
public:
    MupScanner();
};

// globals

extern CharHdl gCharHdl;
extern CmdChar gCmdChar;
extern IllegalChar gIllegalChar;
extern IgnoreChar gIgnoreChar;
extern EndChar gEndChar;
extern GermanQuoteChar gGermanQuoteChar;

extern TextScanner gTextScanner;
extern MathScanner gMathScanner;
extern PlainScanner gPlainScanner;
extern MupScanner gMupScanner;

extern Scanner *gScanner;

#endif // _charhdl_h_

