/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// CmdHdl.cp -- command and environment handlers

#include "ctype.h"
#include "files.h"
#include "charhdl.h"
#include "cmdhdl.h"

// globals 

const int kArgLen = 2*8182;	// max length of command arguments

SymTab<CmdHdl> gCmds;
SymTab<EnvHdl> gEnvs;
SymTab<const char> gDefs;

extern bool putThis(void);

// CmdHdl -- command handler, the base class simply ignores the command

CmdHdl::CmdHdl(int opts, int args) {
    narg = args;
    nopt = opts;
}

// get stream with arg contents

SubStream *
CmdHdl::getArg(Stream *s, Scanner *sc) {
  //    char buff[kArgLen];
  std::string str="";
    gScanner->eatWhite(s);
    char c = s->get();
    StreamPos p = s->pos();
    if (c == '{') {
        if (! sc) sc = gScanner;
        Scanner *as = gScanner;
        gScanner = sc;
        CharHdl *h = sc->getHdl('}');
	sc->setHdl('}', &gEndChar);
	sc->scan(s);	// eats the closing }
	sc->setHdl('}', h);
	gScanner = as;
	//	if (! s->getRange(p, buff, kArgLen))
	//	    s->fatal("argument too long");
	//	buff[strlen(buff)-1] = '\0'; // remove the closing }
	if (! s->getRange(p, str) )
	  s->fatal("Error on reading argument");
    } else {
      if ( c == '\\' ) {
	return gCmdChar.getCmd(s, c);
      } else { // single char arg
      //        buff[0] = c;
      //        buff[1] = '\0';
	str = c;
      }
    }
    //    return new SubStream(s, buff, &p);
    return new SubStream(s, str, &p);
}

// scan arg without processing it

void
CmdHdl::scanArg(Stream *s, Scanner *sc) {
    gScanner->eatWhite(s);
    char c = s->get();
    if (c == '{') {
        if (! sc) sc = gScanner;
        Scanner *as = gScanner;
        gScanner = sc;
        CharHdl *h = sc->getHdl('}');
	sc->setHdl('}', &gEndChar);
	sc->scan(s);	// eats the closing }
	sc->setHdl('}', h);
	gScanner = as;
    }
}

// get stream with optional argument or NULL

SubStream *
CmdHdl::getOptArg(Stream *s, Scanner *sc) {
    gScanner->eatWhite(s);
    char c = s->get();
    if (c == '[') {
      //        char buff[kArgLen];
      std::string str;
        StreamPos p = s->pos();
        if (! sc) sc = gScanner;
        Scanner *as = gScanner;
        gScanner = sc;
        CharHdl *h = sc->getHdl(']');
	sc->setHdl(']', &gEndChar);
	sc->scan(s);	// eats the closing ]
	sc->setHdl(']', h);
	gScanner = as;
	//	  if (! s->getRange(p, buff, kArgLen))
	//	      s->fatal("optional argument too long");
	//	  buff[strlen(buff)-1] = '\0'; // remove the closing ]
	//	return new SubStream(s, buff, &p);
	if (! s->getRange(p, str))
	  s->fatal("error reading optional argument");
	return new SubStream(s, str, &p);
    }
    s->unget(c);	// no opt. arg
    return NULL;
}

// scan optional argument (if exists) without processing it

void
CmdHdl::scanOptArg(Stream *s, Scanner *sc) {
    gScanner->eatWhite(s);
    char c = s->get();
    if (c == '[') {
        if (! sc) sc = gScanner;
        Scanner *as = gScanner;
        gScanner = sc;
        CharHdl *h = sc->getHdl(']');
	sc->setHdl(']', &gEndChar);
	sc->scan(s);	// eats the closing ]
	sc->setHdl(']', h);
	gScanner = as;
    } else {			// no optional arg
        s->unget(c);
    }
}

// translate command, the semantics is defined thru translateArgs

bool
CmdHdl::translate(Stream *s) {
    int i, no;
    bool cont;
    SubStream *a;
    SubStream *opts[9];
    SubStream *args[9];
    // eat white space if no args
    if (nopt == 0 && narg == 0) {
	gScanner->eatWhite(s);
    }
    // get optional args
    for (no = 0, i = 0; i < nopt; i++) {
        a = getOptArg(s);
        if (a) { opts[i] = a; no++; }
        else break;
    }
    for (i = no; i < 9; i++) opts[i] = NULL;
    // get args
    for (i = 0; i < narg; i++) {
        a = getArg(s);
        if (a) args[i] = a;
        else { 
            s->error("argument(s) missing"); 
            return true;
        }
    }
    for (i = narg; i < 9; i++) args[i] = NULL;
    // handle command
    cont = translateArgs(s, opts, no, args);
    // delete args lists
    for (i = 0; i < no; i++) delete opts[i];
    for (i = 0; i < narg; i++) delete args[i];
    return cont;
}

// translate command by handling its args

bool
CmdHdl::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ma[]) {
    return true;
}

// scan over command without processing it

bool
CmdHdl::scan(Stream *s) {
    int i;
    if (nopt == 0 && narg == 0) {
	// eat white space
	gScanner->eatWhite(s);
    } else {
	// scan over args
	for (i = 0; i < nopt; i++) {
	    scanOptArg(s);
	}
	for (i = 0; i < narg; i++) {
	    scanArg(s);
	}
    }
    return true;
}

// BeginCmd, EndCmd -- commands \begin{} and \end{}

const int kEnvLen = 128;	// max length of env. name + 1

// the active environment names are stored in a stack

StrStackLink::StrStackLink(char *c) : StackLink() {
    str = strcpy(new char[strlen(c) + 1], c);
}

StrStackLink::~StrStackLink() {
    delete str;
}

bool
StrStackLink::equal(char *c) {
    return ((strcmp(c, str) == 0) ? true : false);
}
   
Stack<StrStackLink> gCurrEnvName;

static class BeginCmd : public CmdHdl {
public:
    BeginCmd() : CmdHdl(0, 1) {}
    virtual bool translate(Stream *);
    virtual bool scan(Stream *);
} beginCmd;

bool
BeginCmd::translate(Stream *s) {
    SubStream *a = getArg(s);
    EnvHdl *e = gEnvs.top(a->cont());
    if (e) {
	gCurrEnvName.push( new StrStackLink(a->cont()) );
	delete a;
	return e->translate(s);
    }
    s->error("unknown environment: '", a->cont(), "'");
    delete a;
    return true;
}

bool
BeginCmd::scan(Stream *s) {
    SubStream *a = getArg(s);
    EnvHdl *e = gEnvs.top(a->cont());
    if (e) {
	gCurrEnvName.push( new StrStackLink(a->cont()) );
	delete a;
	return e->scan(s);
    }
    delete a;
    return true;
}

static class EndCmd : public CmdHdl {
public:
    EndCmd() : CmdHdl(0, 1) {}
    virtual bool translate(Stream *);
    virtual bool scan(Stream *);
} endCmd;

bool
EndCmd::translate(Stream *s) {
    SubStream *a = getArg(s);
    EnvHdl *e = gEnvs.top(a->cont());
    if (e) {
	if (gCurrEnvName.empty()) {
	    s->error("no corresponding \\begin{", a->cont(), "}");
	} else {
	    StrStackLink *n = gCurrEnvName.top();
	    if (! n->equal(a->cont())) {
	        s->error("no corresponding \\begin{", a->cont(), "}");
	    } else {
	        n = gCurrEnvName.pop();
	        delete n;
	        delete a;
	        return false;
	    }
	}
    } else {
        s->error("unknown environment: '", a->cont(), "'");
    }
    delete a;
    return true;
}

bool
EndCmd::scan(Stream *s) {
    SubStream *a = getArg(s);
    EnvHdl *e = gEnvs.top(a->cont());
    if (e) {
	if (! gCurrEnvName.empty()) {
	    StrStackLink *n = gCurrEnvName.top();
	    if (n->equal(a->cont())) {
	        n = gCurrEnvName.pop();
	        delete n;
	        delete a;
	        return false;
	    }
	}
    }
    delete a;
    return true;
}

// commands which produce accent chars

class AccentHdl : public CmdHdl {
protected:
    char *fAccent;
    char *fHtmlName;
public:
    AccentHdl(char *acc, char *html) : CmdHdl(0, 1) {
	fAccent = acc; fHtmlName = html;
    }
    bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
};

bool
AccentHdl::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[])
{
    if ((strlen(ar[0]->cont()) == 1) &&
	isascii((int) *(ar[0]->cont())) &&
	isalpha((int) *(ar[0]->cont()))) {
	// seems to be a accented char
	Puts("&");
	Puts(ar[0]->cont());
	Puts(fHtmlName);
	Puts(";");
    } else {
	Puts(fAccent);
    }
    return true;
}

// handle commands parsed from definition file

// substitute argument patterns and print to output

static void
subsArgs(Stream *s, char *def, SubStream *opt, SubStream *arg[], int narg) {
    char buff[256], *c;
    int i, tr, lab;
    for (; *def != '\0'; def++) {
        if (*def == '#') {
            def++;
	    if (*def == '=') {
		tr = 0;
		lab = 0;
		def++;
	    } else if (*def == 'L') {
		tr = 0;
		lab = 1;
		def++;
	    } else {
		tr = 1;
		lab = 0;
	    }
            switch (*def) {
            case '\0':
                s->warning("illegal definition: trailing #");
                Putc('#');
		if (! tr) {
		    if (lab) Putc('L'); else Putc('=');
                }
		return;
            case '#':		// ## is subst. by #
		if (! tr) {
		    if (lab) {
			s->warning("illegal definition: formal argument #L#");
			Puts("#L#");
		    } else {
			s->warning("illegal definition: formal argument #=#");
			Puts("#=#");
		    }
		} else {
		    Putc('#');
		}
                break;
            case 'a':		// #a is subst. by opt. arg. no. 1
                if (opt) {
                    opt->reset();
                    if (tr) gScanner->translate(opt);
		    else Puts(opt->cont());
                } else {
                    s->warning("illegal definition: missing optional argument");
		    if (! tr) {
			if (lab) Puts("#La"); else Puts("#=a");
		    } else Puts("#a");
                }
                break;
            case '1':		// #i is subst. by arg. no. i
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                i = (int) (*def - '1');
                if (i < narg && arg[i]) {
                    arg[i]->reset();
                    if (tr) {
			gScanner->translate(arg[i]);
		    } else {
			Puts(arg[i]->cont());
			if (lab) PutLabelAlias(arg[i]->cont());
		    }
                } else { 
                    s->warning("illegal definition: missing argument");
                    Putc('#');
		    if (! tr) {
			if (lab) Putc('L'); else Putc('=');
		    }
		    Putc(*def); 
                }
                break;
                
            case '(':		// #(xyz) is subst. by the value of xyz
		if (! tr) {
		    if (lab)
			s->warning("illegal definition: formal argument #L(..., '=' ignored");
		    else
			s->warning("illegal definition: formal argument #=(..., '=' ignored");
		}
                for (def++, c = buff; (*c = *def) != '\0'; def++, c++) {
                    if (*def == ')') {
                        *c = '\0';
			const char* d = gDefs.top(buff);
                        if (d) {
                            Puts(d);
                        } else {
			  if( 0 == strcmp(buff, "Mthis") ) {
			    Puts("<code>");
			    if (! putThis()) s->error("\\Mthis undefined");
			    Puts("</code>");
			  } else {
                            s->warning("illegal definition: missing macro");
                            Puts("#("); Puts(buff); Putc(')');
			  }
                        }
                        break;
                    }
                    if (c-buff >= 256) { *c = '\0'; break; }
                }
                if (*def != ')') {
                    s->warning("illegal definition: run-away macro");
                    Puts("#("); Puts(buff);
                    if (*def == '\0') return;
                    Putc(*def);
                }
                break;
                
            default:
                Putc('#');
                Putc(*def);
            }
        } else {
            Putc(*def);
        }
    }
}

ParsedCmd::ParsedCmd(int no, int na, char *md, char*od) : 
	CmdHdl(no, na) {
    if (md) def = strcpy(new char[strlen(md)+1], md);
    else def = NULL;
    if (od) optdef = strcpy(new char[strlen(od)+1], od);
    else optdef = NULL;
}

ParsedCmd::~ParsedCmd() {
    if (def) delete def;
    if (optdef) delete optdef;
}

bool
ParsedCmd::translateArgs(Stream *s, SubStream *opts[], int no, SubStream *args[]) {
    if (no == 1 && optdef) subsArgs(s, optdef, opts[0], args, narg);
    else if (def) subsArgs(s, def, opts[0], args, narg);
    return true;
}

// EnvHdl -- environment handler
// the base class simply passes the body thru and ignores the 
// \begin{...} and \end{...} commands

// translate the environment args and body

bool
EnvHdl::translate(Stream *s) {
    int i, no;
    bool cont;
    SubStream *a;
    SubStream *opts[9];
    SubStream *args[9];
    // get optional args
    for (no = 0, i = 0; i < nopt; i++) {
        a = getOptArg(s);
        if (a) { opts[i] = a; no++; }
        else break;
    }
    for (i = no; i < 9; i++) opts[i] = NULL;
    // get args
    for (i = 0; i < narg; i++) {
        a = getArg(s);
        if (a) args[i] = a;
        else { 
            s->error("argument(s) missing"); 
            return true;
        }
    }
    for (i = narg; i < 9; i++) args[i] = NULL;
    // handle command
    cont = translateArgs(s, opts, no, args);
    if (cont) cont = translateBody(s);
    if (cont) cont = translateEnd(s, opts, no, args);
    // delete args lists
    for (i = 0; i < no; i++) delete opts[i];
    for (i = 0; i < narg; i++) delete args[i];
    return cont;
}

// translate the environment args by simply ignoring them

bool
EnvHdl::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ma[]) {
    return true;
}

// translate the environment body by translating it with the active 
// scanner

bool
EnvHdl::translateBody(Stream *s) {
    gScanner->translate(s);
    return true;
}

// translate the environment end by simply ignoring it

bool
EnvHdl::translateEnd(Stream *s, SubStream *oa[], int no, SubStream *ma[]) {
    return true;
}

// scan the environment body, the args are only scanned over

bool
EnvHdl::scan(Stream *s) {
    int i;
    // scan over args
    for (i = 0; i < nopt; i++) {
        scanOptArg(s);
    }
    for (i = 0; i < narg; i++) {
        scanArg(s);
    }
    return scanBody(s);
}

// scan the environment body by scanning it with the active scanner

bool
EnvHdl::scanBody(Stream *s) {
    gScanner->scan(s);
    return true;
}

// handle environments parsed from definition file

ParsedEnv::ParsedEnv(int no, int na, char *md, char *od, char *ed) :
	EnvHdl(no, na) {
    if (md) def = strcpy(new char[strlen(md)+1], md);
    else def = NULL;
    if (od) optdef = strcpy(new char[strlen(od)+1], od);
    else optdef = NULL;
    if (ed) enddef = strcpy(new char[strlen(ed)+1], ed);
    else enddef = NULL;
}

ParsedEnv::~ParsedEnv() {
    if (def) delete def;
    if (optdef) delete optdef;
    if (enddef) delete enddef;
}

bool
ParsedEnv::translateArgs(Stream *s, SubStream *opts[], int no, SubStream *args[]) {
    if (no == 1 && optdef) subsArgs(s, optdef, opts[0], args, narg);
    else if (def) subsArgs(s, def, opts[0], args, narg);
    return true;
}

bool
ParsedEnv::translateEnd(Stream *s, SubStream *opts[], int no, SubStream *args[]) {
    if (enddef) subsArgs(s, enddef, opts[0], args, narg);
    return true;
}

// init the begin and end command handlers

void
initCmds() {
    gCmds.push("begin", &beginCmd);
    gCmds.push("end", &endCmd);
    gCmds.push("\"", new AccentHdl("\"", "uml"));
    gCmds.push("~", new AccentHdl("~", "tilde"));
    gCmds.push("`", new AccentHdl("`", "grave"));
    gCmds.push("'", new AccentHdl("'", "acute"));
    gCmds.push("", new AccentHdl("'", "acute"));
}

// end of file
