/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// CmdHdl.h -- command and environment handlers

#ifndef _cmdhdl_h_
#define _cmdhdl_h_

#include "type.h"
#include "symtab.h"
#include "stream.h"
#include "charhdl.h"

class CmdHdl {
protected:
    int narg, nopt;
    virtual SubStream *getArg(Stream *, Scanner * = NULL);
    virtual SubStream *getOptArg(Stream *, Scanner * = NULL);
    virtual void scanArg(Stream *, Scanner * = NULL);
    virtual void scanOptArg(Stream *, Scanner * = NULL);
public:
    CmdHdl(int, int);
    virtual ~CmdHdl() {};
    virtual bool translate(Stream *);
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool scan(Stream *);
};

class EnvHdl : public CmdHdl {
public:
    EnvHdl(int o, int a) : CmdHdl(o, a) {};
    virtual bool translate(Stream *);
    virtual bool scan(Stream *);
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateBody(Stream *);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool scanBody(Stream *);
};

class ParsedCmd : public CmdHdl {
protected:
    char *def, *optdef;
public:
    ParsedCmd(int, int, char *, char*);
    virtual ~ParsedCmd();
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
};

class ParsedEnv : public EnvHdl {
protected:
    char *def, *optdef, *enddef;
public:
    ParsedEnv(int, int, char *, char *, char*);
    virtual ~ParsedEnv();
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translateEnd(Stream *, SubStream *[], int, SubStream *[]);
};

class StrStackLink : public StackLink {
protected:
    char *str;
public:
    StrStackLink(char *);
    virtual ~StrStackLink();
    bool equal(char *);
};

extern SymTab<CmdHdl> gCmds;
extern SymTab<EnvHdl> gEnvs;
extern SymTab<const char> gDefs;
extern Stack<StrStackLink> gCurrEnvName;

void initCmds();

#endif // _cmdhdl_h_

