/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// files.cpp -- output utils

#include "string.h"
#include "files.h"
#include "mh2html.h"
#include "cmdhdl.h"
#include "basecmds.h"
#include "utils.h"
#include <stack>
#include <algorithm>
#ifndef WIN32
#include <strings.h>
#endif

// Printing to the html file

static FILE *gOutf = NULL;			// output file
static FILE *gAliasf = NULL;
static char *gOutfBaseName = NULL;
static int gOutfNum = 0;
static long gOutfStartPos = 0;

static char *gOutermostBaseName = NULL;

list<string> gFilenames;

class SavedFileInformation {
public:
  FILE *Outf;
  char *OutfBaseName;
  int OutfNum;
  long OutfStartPos;
  SavedFileInformation(FILE *o, char *bn, int n, long pos):
    Outf(o), OutfBaseName(bn), OutfNum(n), OutfStartPos(pos) {}
};

static std::stack<SavedFileInformation> gOuterFiles;

void
OpenHTMLFile(const char *n) {
  if (gOutf) {
    gOuterFiles.push(SavedFileInformation(gOutf,
					  gOutfBaseName,gOutfNum,
					  gOutfStartPos));
  }
    gOutfBaseName = strcpy(new char[strlen(n)+1], n);
    if(NULL== gOutermostBaseName)
      gOutermostBaseName = strdup(n);
    char buff[256];
    strcpy(buff, n);
    strcat(buff, HTML_EXT);
    gOutf = fopen(buff, WRITE_MODE);
    if (gOutf == NULL) 
	fatal("Could not open output file \"", buff, "\"");
    gOutfNum = 0;
    gOutfStartPos = ftell(gOutf);

    if(!gAliasf) {
      strcpy(buff, gOutfBaseName);
      strcat(buff, ".hla");
      gAliasf = fopen(buff, WRITE_MODE);
      if (gAliasf == NULL) 
	fatal("Could not open label alias file \"", buff, "\"");
    }
}

void
CloseHTMLFile() {
  //  if(gAliasf) fclose(gAliasf);
  if(gOutf) fclose(gOutf);
  if(gOutfBaseName) delete[] gOutfBaseName;
  
  if(gOuterFiles.empty()) {
    //    gAliasf       = NULL;
    gOutf         = NULL;
    gOutfBaseName = NULL;
    gOutfNum      = 0;
    gOutfStartPos = 0;
  } else {
    //    gAliasf       = gOuterFiles.top().Aliasf;
    gOutf         = gOuterFiles.top().Outf;
    gOutfBaseName = gOuterFiles.top().OutfBaseName;
    gOutfNum      = gOuterFiles.top().OutfNum;
    gOutfStartPos = gOuterFiles.top().OutfStartPos;

    gOuterFiles.pop();
  }
}

void
OpenNextHTMLFile(Stream *s) {
    if (gOutf == NULL)
	fatal("Internal error [OpenNextHTMLFile]");
    
    if ( ftell(gOutf) == gOutfStartPos )
      return;
    
    PutFooter(s);
    fclose(gOutf);
    gOutfNum++;
    if (gOutfNum > 999)
	fatal("Too many HTML files, choose smaller level (option -l)");
    char buff[256];
    sprintf(buff, "%s%d%s", gOutfBaseName, gOutfNum, HTML_EXT);
    gOutf = fopen(buff, WRITE_MODE);
    if (gOutf == NULL) 
	fatal("Could not open output file \"", buff, "\"");
    PutHeader(s);
}

void
GetContextFileNames(char *prev, char *cont, char *next) {
  string myname = GetHTMLFileName();
  string to_topdir;
  std::list<std::string>::iterator mypos, otherpos;

  sprintf( cont, "index%s", HTML_EXT );

  if ( strncmp ( myname.c_str(), "./", 2) == 0 )
    myname.erase(0,2);

  mypos = std::find( gFilenames.begin(), gFilenames.end(), myname );

  if ( mypos == gFilenames.end() ) {
    //    sprintf(prev, "");
    *prev = '\0';
    //    sprintf(next, "");
    *next = '\0';
    return;
  }

  // find no. of / in filename
  to_topdir = "";
  for( std::string::iterator i= std::find(myname.begin(), 
				myname.end(), '/');
       i != myname.end(); 
       i = std::find( ++i, myname.end(), '/' ) )
    to_topdir += "../";
  
  if ( mypos == gFilenames.begin() )
    //    sprintf( prev, "" );
    *prev = '\0';
  else {
    otherpos = mypos;
    otherpos--;
    sprintf( prev, "%s%s", to_topdir.c_str(), (*otherpos).c_str() );
  }

  otherpos = mypos;
  otherpos++;
  if ( otherpos == gFilenames.end() )
    strcpy ( next, "" );
  else
    sprintf ( next, "%s%s", to_topdir.c_str(), (*otherpos).c_str() );
}

char *
GetHTMLFileName() {
    static char buff[256];
    if(gOutfNum > 0)
      sprintf(buff, "%s%d%s", gOutfBaseName, gOutfNum, HTML_EXT);
    else
      sprintf(buff, "%s%s", gOutfBaseName, HTML_EXT);
    return buff;
}

void
Putc(char c) {
    fputc(c, gOutf);
}

void
Putc2(char c) {
  if ( ftell(gOutf) != gOutfStartPos )
    Putc(c);
}

void 
Puts(const char *s) {
  if(s)
    fputs(s, gOutf);
} 

void 
Puts2(const char *s) {
  if (ftell(gOutf) != gOutfStartPos )
    Puts(s);
}

void 
Putnl() {
    fputc('\n', gOutf);
}

void
PutPar() {
    fputs("<p>\n", gOutf);
    gScanner->donePar();
}

void 
PutVerbc(const char c) {
    switch (c) {
    case '<': Puts("&lt;"); break;
    case '>': Puts("&gt;"); break;
    case '&': Puts("&amp;"); break;
    case '"': Puts("&quot;"); break;
    default: Putc(c);
    }
}

// Printing to the toc file

FILE *
OpenOldTocFile() {
    char buff[256];
    strcpy(buff, gOutermostBaseName);
    strcat(buff, ".htc");
    return fopen(buff, READ_MODE);
}

FILE *gTocf = NULL;

bool 
OpenNewTocFile() {
    if (gTocf) return true;
    char buff[256];
    strcpy(buff, gOutermostBaseName);
    strcat(buff, ".~ht");
    gTocf = fopen(buff, WRITE_MODE);
    return (gTocf ? true : false);
}

void 
PutTocs(const char *s) {
    fputs(s, gTocf);
}

void 
ExitToc(bool doRename) {
    if (gTocf) {
        fclose(gTocf);
        gTocf = NULL;
        char buff[256];
        strcpy(buff, gOutermostBaseName);
        strcat(buff, ".~ht");
        if (doRename) {
            char nnam[256];
            strcpy(nnam, gOutermostBaseName);
            strcat(nnam, ".htc");
            rename(buff, nnam);
        } else
	  remove(buff);
    }
}

// Printing to the contents file

FILE *
OpenOldContFile() {
    char buff[256];
    strcpy(buff, gOutermostBaseName);
    strcat(buff, ".hco");
    return fopen(buff, READ_MODE);
}

FILE *gContf = NULL;

bool 
OpenNewContFile() {
    if (gContf) return true;
    char buff[256];
    strcpy(buff, gOutermostBaseName);
    strcat(buff, ".~hc");
    gContf = fopen(buff, WRITE_MODE);
    return (gContf ? true : false);
}

void 
PutConts(const char *s) {
    fputs(s, gContf);
}

void 
ExitCont(bool doRename) {
    if (gContf) {
        fclose(gContf);
        gContf = NULL;
        char buff[256];
        strcpy(buff, gOutermostBaseName);
        strcat(buff, ".~hc");
        if (doRename) {
            char nnam[256];
            strcpy(nnam, gOutermostBaseName);
            strcat(nnam, ".hco");
            rename(buff, nnam);
        }
        remove(buff);
    }
}

// Printing to the header file

FILE *
OpenOldHeaderFile() {
    char buff[256];
    strcpy(buff, gOutermostBaseName);
    strcat(buff, ".hhd");
    return fopen(buff, READ_MODE);
}

FILE *gHeaderf = NULL;

bool 
OpenNewHeaderFile() {
    if (gHeaderf) return true;
    char buff[256];
    strcpy(buff, gOutermostBaseName);
    strcat(buff, ".~hh");
    gHeaderf = fopen(buff, WRITE_MODE);
    return (gHeaderf ? true : false);
}

void 
PutHeaders(const char *s) {
    fputs(s, gHeaderf);
}

void 
ExitHeader(bool doRename) {
    if (gHeaderf) {
        fclose(gHeaderf);
        gHeaderf = NULL;
        char buff[256];
        strcpy(buff, gOutermostBaseName);
        strcat(buff, ".~hh");
        if (doRename) {
            char nnam[256];
            strcpy(nnam, gOutermostBaseName);
            strcat(nnam, ".hhd");
            rename(buff, nnam);
        } else
	  remove(buff);
    }
}

void
PutLabelAlias(const char* lab) {

    if (gOutfNum == 0) {
	fprintf(gAliasf, "%s#%s\t%s%s#%s\n",
		GetVirtualHTMLFileName(), lab,
		gOutfBaseName, HTML_EXT, lab);
    } else {
	fprintf(gAliasf, "%s#%s\t%s%d%s#%s\n",
		GetVirtualHTMLFileName(), lab,
		gOutfBaseName, gOutfNum, HTML_EXT, lab);
    }
}

// Counter definitions

FILE *
OpenOldAuxFile() {
  char buff[256];
  sprintf(buff, "%s.haux", gOutermostBaseName);
  return fopen(buff, READ_MODE);
}

FILE *gAuxf = NULL;

bool
OpenNewAuxFile() {
  if (gAuxf) return true;
  char buff[256];
  sprintf(buff, "%s.~haux", gOutermostBaseName);
  gAuxf = fopen(buff, WRITE_MODE);
  return (gAuxf ? true : false);
}

void
PutAux(const char *s) {
  OpenNewAuxFile();
  fputs(s, gAuxf);
}

void PutLabel(const string &label, int num) {
  // since our installation does not have stringstream yet (sigh), ...
  char buff[256];
  sprintf(buff, "\\makelabel{%s}{%d}\n", label.c_str(), num);
  PutAux(buff);
}

void PutLocalLabel(const string& label, int num) {
  // since our installation does not have stringstream yet (sigh), ...
  char buff[256];
  sprintf(buff, "\\makelocallabel{%s}{%s}{%d}\n", 
	  GetHTMLFileName(), label.c_str(), num);
  PutAux(buff);
}

void
ExitAux(bool doRename) {
  if (gAuxf) {
    fclose(gAuxf);
    gAuxf = NULL;
    char buff[256];
    sprintf(buff, "%s.~haux", gOutermostBaseName);
    if (doRename) {
      char nnam[256];
      sprintf(nnam, "%s.haux", gOutermostBaseName);
      rename(buff, nnam);
    } else
      remove(buff);
  }
}

void
ReadAux(Stream *s) {
  FILE *oldAuxf;

  oldAuxf = OpenOldAuxFile();
  if(NULL==oldAuxf) {
    s->warning("no aux file present, run again to fix crossreferences");
    return;
  }

  {
    localScanner _(&gTextScanner);
    FileStream inStream(oldAuxf, ".haux");
    gScanner->translate(&inStream);
  }
}
// put context navigation menu

static void
putContextMenu() {
    const char *c;
    char prev[256], cont[256], next[256];
    GetContextFileNames(prev, cont, next);
    Puts("<p align=right>");
    gScanner->donePar();
    if (strlen(prev) != 0) {
	Puts("<a href=\"");
	Puts(prev);
	Puts("\">");
	if ( (c = gDefs.top("prevpage")) ) Puts(c); else Puts("[Previous]");
	Puts("</a> ");
    }
    if (strlen(next) != 0) {
	Puts("<a href=\"");
	Puts(next);
	Puts("\">");
	if ( (c = gDefs.top("nextpage")) ) Puts(c); else Puts("[Next]");
	Puts("</a> ");
    }
    if (strlen(cont) != 0) {
	Puts("<a href=\"");
	Puts(cont);
	Puts("\">");
	if ( (c = gDefs.top("contpage")) ) Puts(c); else Puts("[Contents]");
	Puts("</a> ");
    }
    if ( (c = gDefs.top("indexpage")) ) {
	Puts("<a href=\"index"); Puts(HTML_EXT); Puts("\">");
	Puts(c);
	Puts("</a> ");
    }
    if ( (c = gDefs.top("findpage")) ) {
	Puts("<a href=\"finder"); Puts(HTML_EXT); Puts("\">");
	Puts(c);
	Puts("</a> ");
    }
    Puts("</p>\n");
    gScanner->markPar();
}

// the following handler is active when reading the *.hhd file

static class TitleHeaderCmd : public CmdHdl {
public:
    TitleHeaderCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} titleHeaderCmd;

bool
TitleHeaderCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    Puts("<title>");
    gScanner->translate(a[0]);
    Puts("</title>\n");
    return true;
}

// put page header

void
PutHeader(Stream *s) {
    FILE *oldHeaderf;
    const char* c;

    Puts("<!doctype html public \"-//IETF//DTD HTML 3.2//EN\">\n");
    Puts("<html>\n<head>\n");

    // put header parts from *.hhd file
    if ( (oldHeaderf = OpenOldHeaderFile()) ) {
        // redefine \title locally
        gCmds.push("title", &titleHeaderCmd);
        FileStream inStream(oldHeaderf, ".hhd");
        Scanner *sc = gScanner;
        gScanner = &gTextScanner;
        gScanner->translate(&inStream);
        gScanner = sc;
        gCmds.pop("title");
    } else {
        s->warning("no header file present, run again to fix title");
    }

    // put header macro
    if ( (c = gDefs.top("header")) ) Puts(c);
    Puts("\n</head>\n<body BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\">\n");
    if ( (c = gDefs.top("basefont")) ) Puts(c);

    // put navigation menu
    putContextMenu();

    // To detect whether content has been written
    gOutfStartPos = ftell(gOutf);
}

// put page footer

void 
PutFooter(Stream *s) {
    const char* c;

    // put navigation menu
    putContextMenu();

    // put footer macro
    if ( (c = gDefs.top("footer")) ) Puts(c);
    Puts("\n</body>\n</html>\n");
}


// end of file
