/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// HyTxCmds.cp -- HyTeX-commands

#include <string.h>
#include "mh2html.h"
#include "files.h"
#include "charhdl.h"
#include "hytxcmds.h"

// nodepageref* command
// the first arg may contain $'s interpreted specially

static class NodepagerefSCmd : public CmdHdl {
public:
    NodepagerefSCmd() : CmdHdl(0, 2) {}
    virtual bool translate(Stream *);
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} nodepagerefSCmd;

bool
NodepagerefSCmd::translate(Stream *s) {
    int i;
    bool cont;
    SubStream *opts[9];
    SubStream *args[9];
    for (i = 0; i < 9; i++) opts[i] = NULL;
    // get args
    if (! (args[0] = getArg(s, &gPlainScanner))) { 
        s->error("1st argument missing"); 
        return true;
    }
    if (! (args[1] = getArg(s))) { 
        s->error("2nd argument missing"); 
        return true;
    }
    for (i = 2; i < 9; i++) args[i] = NULL;
    // handle command
    cont = translateArgs(s, opts, 0, args);
    // delete args lists
    delete args[0];
    delete args[1];
    return cont;
}

bool
NodepagerefSCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    char buff[256];
    if (! a[0]->copy(buff, 255)) {
        a[0]->error("label too long");
        gScanner->translate(a[1]);
    } else {
        char *c;
        Puts("[<b><a href=\"");
        // an argument lab$file represents an html-label file#lab
        if ( (c = strchr(buff, '$')) ) {
            *c = '\0'; Puts(c+1); Puts(HTML_EXT); Puts("#"); Puts(buff);
        } else {
	    Puts(GetVirtualHTMLFileName());
            Putc('#'); Puts(buff);
        }
        Puts("\">");
        gScanner->translate(a[1]);
        Puts("</a></b>] ");
    }
    return true;
}

// init the command handlers

void
initHyTeXCmds() {
    gCmds.push("nodepageref*", &nodepagerefSCmd);
}

// end of file
