/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */

// List.h -- classes for linked lists

#ifndef _list_h_
#define _list_h_

// the following must be a superclass for elements of intrusive
// liked lists

class ListLink {
    friend class ListBase;
    ListLink *fNext;
public:
    ListLink() { fNext = NULL; }
    virtual ~ListLink() { if (fNext) delete fNext; }
};

// the element links for non-intrusive lists

template<class T>
class ElemLink : public ListLink {
    T *fElem;
public:
    ElemLink(T *e) : ListLink() { fElem = e; }
    virtual ~ElemLink() {} // may be overloaded to allow deletion of
    			   // elements
    T *elem() { return fElem; }
};

// helper class for lists

class ListBase {
    ListLink *fFirst;
    ListLink *fAct;
public:
    ListBase() { 
        fFirst = NULL;
        fAct = NULL;
    }
    virtual ~ListBase() {
        if (fFirst) delete fFirst;
    }
    ListLink* head() { 
        if (fFirst) fAct = fFirst->fNext;
        else fAct = NULL;
        return fFirst;
    }
    void insert(ListLink *e) {
        e->fNext = fFirst;
        fFirst = e;
    }
    ListLink* next() {
        if (fAct) {
            ListLink* l = fAct;
            fAct = l->fNext;
            return l;
        }
        return NULL;
    }
    void clear() {
        delete fFirst; // removes all links
        fFirst = NULL;
        fAct = NULL;
    }
};

// an intrusive list class

template<class T>
class IList : public ListBase {
public:
    IList() : ListBase() {}
    void insert(T *e) {
        ListBase::insert((ListLink *) e);
    }
    T* head() { 
        return (T *) ListBase::head();
    }
    T* next() {
        return (T *) ListBase::next();
    }
};

// an non-intrusive list class

template<class T>
class List : public ListBase {
public:
    List() : ListBase() {}
    void insert(T *e) {
        ElemLink<T> *l = new ElemLink<T>(e);
        ListBase::insert((ListLink *) l);
    }
    T *head() { 
        ElemLink<T> *l = (ElemLink<T> *) ListBase::head();
        return (T *) ((l) ? l->elem() : NULL);
    }
    T *next() {
        ElemLink<T> *l = (ElemLink<T> *) ListBase::next();
        return (T *) ((l) ? l->elem() : NULL);
    }
};

#endif // _list_h_
