/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */


#include <stdlib.h>
#include <string.h>
#ifdef WINDOWS
#include <fcntl.h>
#include <locale.h>
#endif
#ifdef MW
#include <console.h>
#endif
#include "mh2html.h"
#include "files.h"
#include "stream.h"
#include "charhdl.h"
#include "basecmds.h"
#include "baseenvs.h"
#include "hytxcmds.h"
#include "mhlpcmds.h"
#include "mhlpenvs.h"
#include "cmdfile.h"

#define SYS_DEFS "mh2html.def"

static char gVirtualHTMLFileName[256];

const char *
GetVirtualHTMLFileName() {
    return gVirtualHTMLFileName;
}

bool gIe4 = false;

void usage(char *);

void
usage(char *c) {
    printf("usage: %s [-h] [-d #] [-o #] [-l #] [-D #]* <tex-file>\n\n", c);
    printf("       -h	this help\n");
    printf("       -v	version info\n");
    printf("       -d #	definition file\n");
    printf("       -o #	base name of output file\n");
    printf("       -l #	file break level [1-6]\n");
    printf("       -D #	macro definition\n\n");
    exit(1);
}

main(int argc, char *argv[]) {
    FILE *inf, *def;
    int bl, lev, i;
    char *bnam = NULL, *dfile = NULL, *dpath, *ofile = NULL;

#ifdef WINDOWS
    _fmode = _O_BINARY;
    setlocale(LC_ALL, "C");
#endif

    // open input and output files
#ifdef MW
    argc = ccommand(&argv);
#endif

#ifdef MW
    dpath = ":";
#else
    dpath = getenv("MH2HTML_PATH");
#endif

    for (i = 1; i < argc; i++) {
	if (strcmp(argv[i], "-h") == 0) {
	    usage(argv[0]);
	} else if (strcmp(argv[i], "-v") == 0) {
            fprintf(stderr, "mh2html V 2.0\n");
            exit(0);
	} else if (strcmp(argv[i], "-d") == 0) {
	    i++;
	    if (i >= argc) usage(argv[0]);
	    dfile = argv[i];
	} else if (strcmp(argv[i], "-o") == 0) {
	    i++;
	    if (i >= argc) usage(argv[0]);
	    ofile = argv[i];
	} else if (strcmp(argv[i], "-l") == 0) {
	    i++;
	    if (i >= argc) usage(argv[0]);
	    lev = atoi(argv[i]);
	    if (lev <= 0 || lev > 6) usage(argv[0]);
	    SetBreakLevel(lev);
	} else if (strcmp(argv[i], "-D") == 0) {
	    i++;
	    if (i >= argc) usage(argv[0]);
	    gDefs.push(argv[i], "1");
	} else break;
    }

    if (i+1 != argc) usage(argv[0]);

    bl = strlen(argv[i]);
    bnam = new char[bl+5];
    strcpy(bnam, argv[i]);
    /* add .tex extension if not given */
    if (bl > 4) {
	if (strcmp(bnam+bl-4, ".tex")) {
	    strcat(bnam, ".tex");
	} else {
	    bl -= 4;
	}
    } else {
	strcat(bnam, ".tex");
    }

    // switch language
    if (gDefs.top("german"))
	gLanguage = kGerman;
    else
	gLanguage = kEnglish;

    // Internet Explorer 4.x output?
    if (gDefs.top("ie4"))
        gIe4 = true;

    if (! dfile) {
        if (dpath) {
	    dfile = new char[strlen(dpath)+strlen(SYS_DEFS)+2];
	    strcpy(dfile, dpath);
	    if (*(dfile+strlen(dpath)-1) != PATH_SEP_CHAR)
	        strcat(dfile, PATH_SEP_STR);
        } else {
	    dfile = new char[strlen(SYS_DEFS)+1];
            *dfile = '\0';
        }
	strcat(dfile, SYS_DEFS);
    }
    
    inf = fopen(bnam, READ_MODE);
    if (! inf) fatal("Could not open input file \"", bnam, "\"");
    // create the main stream
    FileStream inStream(inf, bnam);

    bnam[bl] = '\0';

    // create name of 'virtual output html file' and define macro
    strcpy(gVirtualHTMLFileName, bnam);
    strcat(gVirtualHTMLFileName, HTML_EXT);
    gDefs.push("thisfile", gVirtualHTMLFileName);

    if (ofile == NULL) ofile = bnam;
    OpenHTMLFile(ofile);

    initCmds();
    initBasicCmds();
    initBasicEnvs();
    initHyTeXCmds();
    initMhelpCmds();
    initMhelpEnvs();

    // read command definitions
    def = fopen(dfile, READ_MODE);
    if (! def)
	fatal("Could not open command definitions file \"", dfile, "\"");
    FileStream defStream(def, dfile);
    initCmdsFromFile(&defStream, dfile);

    ReadAux(&inStream);

    PutHeader(&inStream);
    PutContents(&inStream);
    gScanner->translate(&inStream);
    PutFooter(&inStream);

    ExitHeader(true);
    ExitCont(true);
    ExitToc(true);
    ExitAux(true);
    return(0);
}

// end of file
