/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */


// MhlpCmds.cp -- special Mhelp.sty commands

#include "files.h"
#include "mh2html.h"
#include "charhdl.h"
#include "basecmds.h"
#include "baseenvs.h"
#include "mhlpcmds.h"
#include "utils.h"
#include "counter.h"
#include <ctype.h>

SubStream gMpack;
SubStream gMname;
SubStream gMnameLabel;
SubStream gAuthor;
SubStream gIncpath;
ParaType gParaType = kPara;
bool gPutVersion = false;
bool gPageButtons = false;
bool gInSyntax = false;
Language gLanguage = kEnglish;
FILE *gOldTocf = NULL;

#ifdef __NON_SPLITTING_REF__

// putRef -- put anchor with text t
static void
putRef(SubStream *a, SubStream *t) {
    // if a is non-null it defines the ref, otherwise the label 
    // is a local one and t defines the ref.
    Puts("<a href=\"");
    if (a) {
	char *c, *tr;
	tr = strchr(a->cont(), ',');
	if (tr == NULL) {
	    // local ref
	  /*	    Puts(GetVirtualHTMLFileName());
	   *	    Putc('#'); */
	    Puts(a->cont());
	    Puts(HTML_EXT);
	} else {
	    // put virtual file name
	    Puts("../");
	    for (c = a->cont(); c < tr; c++)
		Putc(*c);
	    Puts("/");
	    c = tr+1;
	    if (*c == 0) {
		// put default ref
		Puts(t->cont());
		t->reset();
	    } else {
		Puts(c);
	    }
	    Puts(HTML_EXT);
	}
	a->reset();
    } else {
      /*	Puts(GetVirtualHTMLFileName());
       *	Putc('#');
       *	Puts(t->cont());
       *	t->reset(); */
      Puts(t->cont());
      t->reset();
      Puts(HTML_EXT);
    }
    Puts("\">");
    gScanner->translate(t);
    Puts("</a>");
}

#else

// putRef -- put anchor with text t
static void
putRef(SubStream *a, SubStream *t) {
    // if a is non-null it defines the ref, otherwise the label 
    // defines the ref.
    char *c, *tr;
    Puts("<a href=\"");
    if (a) {
	tr = strchr(a->cont(), ',');
	if (tr == NULL) {
	    // local ref
	  /*	    Puts(GetVirtualHTMLFileName());
	   *	    Putc('#'); */
	    Puts(a->cont());
	    Puts(HTML_EXT);
	} else {
	    // put virtual file name
	    Puts("../");
	    for (c = a->cont(); c != tr; c++)
		Putc(*c);
	    Putc('/');
	    c = tr+1;
	    if (*c == 0) {
		// put default ref
		Puts(t->cont());
		t->reset();
	    } else {
		Puts(c);
	    }
	    Puts(HTML_EXT);
	}
	a->reset();
    } else {
        // split at ::
        char* tr = strchr(t->cont(), ':');
        if (tr != NULL && *(tr+1) == ':') {
	    // put virtual file name
	    Puts("../");
	    for (c = t->cont(); isspace(*c); ++c);
            for ( ; c != tr; c++)
                Putc(*c);
	    Putc('/');
            Puts(tr+2);
	    Puts(HTML_EXT);
        } else {
	    Puts("../stdlib/");
	    Puts(t->cont());
	    Puts(HTML_EXT);
        }
	t->reset();
    }
    Puts("\">");
    gScanner->translate(t);
    Puts("</a>");
}

#endif

// getDef -- return value of macro
static const char *
getDef(const char *name) {
    const char *v = gDefs.top(name);
    if (v == NULL) v = "";
    return v;
}

void putLangPara(SubStream *s)
{
    switch (gParaType) {
    case kNone: break;
    case kPara: Puts("<p>"); break;
    case kItPara: Puts("<p class=\"pit\">"); break;
    case kItem: Puts("<li>"); break;
    case kTableEntry: Puts("<td valign=\"top\">"); break;
    case kLastTableEntry: Puts("<td valign=\"top\">"); break;
    }
    gScanner->donePar();
    gScanner->translate(s);
    switch (gParaType) {
    case kNone: break;
    case kPara: Puts("</p>"); break;
    case kItPara: Puts("</p>"); break;
    case kItem: Puts("</li>"); break;
    case kTableEntry: Puts("</td>"); break;
    case kLastTableEntry: Puts("</td></tr>"); break;
    }
}

// de is german text

static class DeCmd : public CmdHdl {
public:
    DeCmd() : CmdHdl(1, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} deCmd;

bool
DeCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    if (gLanguage == kGerman) {
        putLangPara(a[0]);
    }
    return true;
}

// eng is english text

static class EngCmd : public CmdHdl {
public:
  EngCmd() : CmdHdl(1, 1) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} engCmd;

bool
EngCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    if (gLanguage == kEnglish) {
        putLangPara(a[0]);
    }
    return true;
}

// Minput is a special form of the \input comand

static class MinputCmd : public CmdHdl {
public:
    MinputCmd() : CmdHdl(0, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} minputCmd;

bool
MinputCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    FILE *inf;
    char buff[256];
    if (! gIncpath.copy(buff,250) ) {
        a[0]->error("path name too long, \\Minput ignored");
        return true;
    }
    if(strlen(buff)>0)
      strcat(buff, "/");
    int n = strlen(buff);
    // get directory name
    if (! a[0]->copy(buff+n, 250-n)) {
        a[0]->error("path name too long, \\Minput ignored");
        return true;
    }
    // get file name
    strcat(buff,"/");
    n = strlen(buff);
    if (! a[1]->copy(buff+n, 250-n)) {
        a[1]->error("path name too long, \\Minput ignored");
        return true;
    }
    strcat(buff, ".tex");
    // open and translate file
    if ((inf = fopen(buff, READ_MODE)) == NULL) {
        a[1]->error("could not open \"",buff,"\", \\Minput ignored");
        return true;
    }
    FileStream inStream(inf, buff);
    strcpy(buff, "./");
    if(!gMpack.empty()) {
      gMpack.copy(buff+2,240);
      strcat(buff,"/");
    }
    n = strlen(buff);
    if (! a[1]->copy(buff+n, 250-n)) {
        a[1]->error("path name too long, \\Minput ignored");
        return true;
    }
    OpenHTMLFile(buff); // stores the previous file
    PutHeader(s);
    {
      localScanner _(&gTextScanner);
      gScanner->donePar();
      gScanner->translate(&inStream);
    }
    PutFooter(s);
    CloseHTMLFile();
    return true;
}

static class MincpathCmd : public CmdHdl {
public:
  MincpathCmd() : CmdHdl(0, 1) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mincpathCmd;

bool
MincpathCmd::translateArgs(Stream *s, SubStream *[], int, SubStream *a[]) {
  gIncpath = a[0];
  return true;
}

// Mname returns the name of the current function/variable/domain

class MnameCmd : public CmdHdl {
public:
    MnameCmd() : CmdHdl(0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mnameCmd;

bool
MnameCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    if (! gMname.empty()) { 
	Puts("<code>"); 
	gScanner->translate(&gMname);
	gMname.reset();
	Puts("</code>");
    }
    else s->error("\\Mname undefined");
    return true;
}

// Mthis gives the (full) name of the current object, including package,
// in the form pack::name.

bool
putThis() {
    if (! gMpack.empty() 
	&& strcmp(gMpack.cont(),"stdlib") != 0
	&& strcmp(gMpack.cont(),"datatypes") != 0
	) { 
        if (! gMname.empty()) {
            gScanner->translate(&gMpack);
	    gMpack.reset();
	    Puts("::"); 
	    gScanner->translate(&gMname);
	    gMname.reset();
        } else {
            gScanner->translate(&gMpack);
	    gMpack.reset();
        }
    } else if (! gMname.empty()) {
	gScanner->translate(&gMname);
	gMname.reset();
    } else {
        Puts("***UNKNOWN***");
        return false;
    }
    return true;
}

class MthisCmd : public CmdHdl {
public:
    MthisCmd() : CmdHdl(0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mthisCmd;

bool
MthisCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    Puts("<code>");
    if (! putThis()) s->error("\\Mthis undefined");
    Puts("</code>");
    return true;
}

// Mcall is a function call with function \Mthis

static class McallCmd : public CmdHdl {
public:
    McallCmd() : CmdHdl(1, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mcallCmd;

bool
McallCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
  if (no > 0) {
    Puts("<br><code>");
    gScanner->translate(oa[0]);
    Putc('(');
    gScanner->translate(ar[0]);
    Puts(")</code>\n");
  } else {
    if (! gMname.empty() ) { 
      Puts("<br><code>");
      if (! gMpack.empty() 
	  &&  strcmp ( gMpack.cont(), "stdlib" ) != 0
	  &&  strcmp ( gMpack.cont(), "datatypes" ) != 0
	  ) {
	gScanner->translate(&gMpack);
	gMpack.reset();
	Puts("::");
      }
      gScanner->translate(&gMname);
      gMname.reset();
      Putc('(');
      gScanner->translate(ar[0]);
      Puts(")</code>\n");
    }
    else s->error("\\Mname undefined");
  }
  return true;
}

// Mkernel states that 'this' is a kernel function

class MkernelCmd : public CmdHdl {
public:
    MkernelCmd() : CmdHdl(1, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mkernelCmd;

bool
MkernelCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    Puts("<li><code>");
    if(0==no) {
      if (! putThis()) s->error("\\Mthis undefined");
    } else {
      localScanner _(&gMupScanner);
      gMupScanner.translate(oa[0]);
    }
    Puts("</code> ");
    Puts(getDef("kernelfunc"));
    Puts(".</li>");
    return true;
}

// Margs contains the syntax of the args of a function

class MargsCmd : public CmdHdl {
public:
    MargsCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} margsCmd;

bool
MargsCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    Puts("<code>");
    if (! putThis()) s->error("\\Mthis undefined");
    Puts("(</code>");
    gScanner->translate(ar[0]);
    Puts("<code>)</code>");
    if (gInSyntax) Puts("<br>");
    return true;
}

// Mup is the name of a MuPAD object to be referred to

class MupCmd : public CmdHdl {
public:
  MupCmd() : CmdHdl(1, 1) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
  virtual bool translate(Stream *);
} mupCmd;

bool
MupCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    localScanner _(&gMupScanner);

    Puts("<code>");
    putRef(((no == 1) ? oa[0] : (SubStream*)NULL), a[0]);
    Puts("</code>");
    return true;
}

bool MupCmd::translate(Stream *s) {
  localScanner _(&gMupScanner);

  return CmdHdl::translate(s);
}

// Mupc is the name of a MuPAD object with arguments to be referred to

class MupcCmd : public CmdHdl {
public:
    MupcCmd() : CmdHdl(1, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
    virtual bool translate(Stream *);
} mupcCmd;

bool
MupcCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    localScanner _(&gMupScanner);

    Puts("<code>");
    putRef(((no == 1) ? oa[0] : (SubStream*)NULL), a[0]);
    //    Puts("(</code>");
    Puts("(");
    gScanner->translate(a[1]);
    //    Puts("<code>)</code>");
    Puts(")</code>");
    return true;
}

bool MupcCmd::translate(Stream *s) {
  localScanner _(&gMupScanner);

  return CmdHdl::translate(s);
}

// ----------------------------------------------------------------------

class MupToDoCmd : public CmdHdl {
public:
  MupToDoCmd() : CmdHdl(1, 1) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
  virtual bool translate(Stream *);
} mupToDoCmd;

bool
MupToDoCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    localScanner _(&gMupScanner);

    Puts("<code>");
    gScanner->translate(a[0]);
    Puts("</code>");
    return true;
}

bool MupToDoCmd::translate(Stream *s) {
  localScanner _(&gMupScanner);

  return CmdHdl::translate(s);
}

// ----------------------------------------------------------------------

class MslotCmd : public CmdHdl {
public:
  MslotCmd() : CmdHdl(1, 1) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
  virtual bool translate(Stream *);
} mslotCmd;

bool
MslotCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    localScanner _(&gMupScanner);

    Puts("<code>&quot;");
    gScanner->translate(a[0]);
    Puts("&quot;</code>");
    return true;
}

bool MslotCmd::translate(Stream *s) {
  localScanner _(&gMupScanner);

  return CmdHdl::translate(s);
}

// ----------------------------------------------------------------------

class MexCmd : public CmdHdl {
public:
  MexCmd() : CmdHdl(0, 1) {}
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
  virtual bool translate(Stream *);
} mexCmd;

bool
MexCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    localScanner _(&gMupScanner);

    Puts("<code>");
    gScanner->translate(a[0]);
    Puts("</code>");
    return true;
}

bool MexCmd::translate(Stream *s) {
  localScanner _(&gMupScanner);

  return CmdHdl::translate(s);
}


// Mentryref is a refrence to a domain entry

class MentryrefCmd : public CmdHdl {
public:
    MentryrefCmd() : CmdHdl(1, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mentryrefCmd;

bool
MentryrefCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    Puts("<tr><td><code>");
    gScanner->translate(a[0]);
    Puts("</code> <td>-- <code>");
    putRef(((no == 1) ? oa[0] : (SubStream*)NULL), a[1]);
    Puts("</code><br>");
    return true;
}

// Mlanggerman enables output of german envs. Mgerman

class MlanggermanCmd : public CmdHdl {
public:
    MlanggermanCmd() : CmdHdl(0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mlanggermanCmd;

bool
MlanggermanCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    gLanguage = kGerman;
    return true;
}

// Mlangenglish enables output of german envs. Menglish

class MlangenglishCmd : public CmdHdl {
public:
    MlangenglishCmd() : CmdHdl(0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mlangenglishCmd;

bool
MlangenglishCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    gLanguage = kEnglish;
    return true;
}

// Mhref refers to a local or external hyper-node

class MhrefCmd : public CmdHdl {
public:
    MhrefCmd() : CmdHdl(1, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mhrefCmd;

bool
MhrefCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
  if(no==1) {
    putRef(oa[0], ar[0]);
  } else {
    std::pair<std::string,std::string> target(GetHTMLFileName(), ar[0]->cont());
    char buff[128];
    if (gLocalLables.find(target) != gLocalLables.end()) {
      sprintf(buff, "%d", (*(gLocalLables.find(target))).second);      
    } else {
      sprintf(buff, "[>>]");
    }
    SubStream text(ar[0], buff);
    putRef(ar[0], &text);
  }
  return true;
}

// Mpacktitle creates a package title

class MpacktitleCmd : public CmdHdl {
public:
    MpacktitleCmd() : CmdHdl(1, 5) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mpacktitleCmd;

bool
MpacktitleCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    const char *c;
    // printer title into header file
    if (OpenNewHeaderFile()) {
        PutHeaders("\\title{");
	PutHeaders(ar[1]->cont());
	PutHeaders("}\n");
    } else s->error("could not open header file");

    // put address
    Puts("<address>");
    if ((c = getDef("libdoc"))) {
	Puts(c);
    } else {
	Puts("MuPAD Library Document No");
    }
    Puts(". ");
    gScanner->translate(ar[0]);
    if (no == 1) {
	Puts(" [V. ");
	gScanner->translate(oa[0]);
	Puts("]");
    }
    Puts("<br>\n");
    gScanner->translate(ar[2]);
    Puts("<br>\n");
    gScanner->translate(ar[3]);
    Puts("</address>\n");

    return true;
}

// Mtocline creates a reference for the contents section

class MtoclineCmd : public CmdHdl {
public:
    MtoclineCmd() : CmdHdl(1, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mtoclineCmd;

bool
MtoclineCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    Puts("<dt><code>");
    putRef(((no == 1) ? oa[0] : (SubStream*)NULL), ar[0]);
    Puts("</code><dd>");
    gScanner->translate(ar[1]);
    return true;
}

// Maddtoconts adds an entry to the toc file

class MaddtocontsCmd : public CmdHdl {
public:
    MaddtocontsCmd() : CmdHdl(1, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} maddtocontsCmd;

bool
MaddtocontsCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    // print toc line
    if (OpenNewTocFile()) {
        PutTocs("\\Mtocline");
        if (no == 1) {
            PutTocs("[");
            PutTocs(oa[0]->cont());
            PutTocs("]");
        }
        PutTocs("{");
        PutTocs(ar[0]->cont());
        PutTocs("}{");
        PutTocs(ar[1]->cont());
        PutTocs("}\n");
    } else s->error("could not open toc file");
    return true;
}

// Mcontents includes the toc lines of the help pages

class McontentsCmd : public CmdHdl {
public:
    McontentsCmd() : CmdHdl(0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mcontentsCmd;

bool
McontentsCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    if ((gOldTocf = OpenOldTocFile())) {
	Puts("<dl>\n");
	FileStream inStream(gOldTocf, ".htc");
	localScanner _(&gTextScanner);

	gScanner->translate(&inStream);
	Puts("</dl>\n");
    } else {
        s->warning("no toc file present, run again to fix references");
        Puts("<blockquote>*** NO TOC FILE (.htc) ***</blockquote>\n");
    }
    return true;
}

// Mversion switches version output on/off

class MversionCmd : public CmdHdl {
public:
    MversionCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mversionCmd;

bool
MversionCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    if (strcmp(ar[0]->cont(), "off") == 0) gPutVersion = false;
    else if (strcmp(ar[0]->cont(), "on") == 0) gPutVersion = true;
    else s->error("illegal argument [\\Mversion]");
    return true;
}

// Moutput switches between output options, currently only 'ascii' is handled
// specially

class MoutputCmd : public CmdHdl {
public:
    MoutputCmd() : CmdHdl(0, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} moutputCmd;

bool
MoutputCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    if (strcmp(ar[0]->cont(), "hytex") == 0) gPageButtons = true;
    else gPageButtons = false;
    return true;
}

// Mmenuhref defines a node page ref

class MmenuhrefCmd : public CmdHdl {
public:
    MmenuhrefCmd() : CmdHdl(1, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mmenuhrefCmd;

bool
MmenuhrefCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    if (! gPageButtons) return true;
    Puts("[<b>");
    putRef(((no == 1) ? oa[0] : (SubStream*)NULL), ar[0]);
    Puts("</b>] ");
    return true;
}

// Mbookbuttons is a utility to define some standard node page refs

class MbookbuttonsCmd : public CmdHdl {
public:
    MbookbuttonsCmd() : CmdHdl(0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mbookbuttonsCmd;

bool
MbookbuttonsCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    if (! gPageButtons) return true;
    Puts("[<b><a href=\"man_eng.html#tableofcontents\">");
    Puts(getDef("tableofcontents"));
    Puts("</a></b>] [<b><a href=\"man_eng.html#help-index\">");
    Puts(getDef("helpindex"));
    Puts("</a></b>] [<b><a href=\"man_eng.html#index\">");
    Puts(getDef("index"));
    Puts("</a></b>] [<b><a href=\"man_eng.html#news\">");
    Puts(getDef("news"));
    Puts("</a></b>] ");
    return true;
}

// Mconts includes only the function names of the toc lines of the help pages
// Mtocline is redefined locally to return only the first arg
// MInfo is redefined locally and translates its boy only

class McontsToclineCmd : public CmdHdl {
public:
    McontsToclineCmd() : CmdHdl(1, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mcontsToclineCmd;

bool
McontsToclineCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    Puts("<code>");
    putRef(((no == 1) ? oa[0] : (SubStream*)NULL), ar[0]);
    Puts("</code>");
    return true;
}

class McontsCmd : public CmdHdl {
public:
    McontsCmd() : CmdHdl(0, 0) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mcontsCmd;

bool
McontsCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    if ((gOldTocf = OpenOldTocFile())) {
        // redefine \Mtocline and MInfo locally
        gCmds.push("Mtocline", &mcontsToclineCmd);
        gEnvs.push("MInfo", &gNullEnv);
	Puts("<blockquote>\n");
	FileStream inStream(gOldTocf, ".htc");
	localScanner _(&gTextScanner);

	gScanner->translate(&inStream);
	Puts("</blockquote>\n");
	gCmds.pop("Mtocline");
        gEnvs.pop("MInfo");
    } else {
        s->warning("no toc file present, run again to fix references");
        Puts("<blockquote>*** NO TOC FILE (.htc) ***</blockquote>\n");
    }
    return true;
}

// OptionHdl is an option declared via MDeclareOption

class OptionHdl : public CmdHdl {
public:
    OptionHdl(const char* v) : CmdHdl(0, 0) { 
        val = strcpy(new char[strlen(v)+1], v); 
    }
    virtual ~OptionHdl() { delete val; }
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
protected:
    char* val;
};

bool
OptionHdl::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    Puts("<i class=\"Option\">");
    Puts(val);
    Puts("</i>");
    return true;
}

// MDeclareOption declares a new option

static class MDeclareOptionHdl : public CmdHdl {
public:
    MDeclareOptionHdl() : CmdHdl(0, 2) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mDeclareOptionHdl;

bool
MDeclareOptionHdl::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    char* c = ar[0]->cont();
    if (*c != '\\') {
        ar[0]->error("illegal option declaration");
        return false;
    }
    gCmds.push(c+1, new OptionHdl(ar[1]->cont()));
    return true;
}

// MOverloading lists overloadable args

static class MOverloadingCmd : public CmdHdl {
public:
    MOverloadingCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mOverloadingCmd;

bool
MOverloadingCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    // '<h3>#(overloadable):</h3><p><code>#1</code></p>'
    Puts("<h3>");
    Puts(getDef("overloadable"));
    Puts(":</h3><p>\n<code>");
    gScanner->donePar();

    char* a = ar[0]->cont();
    char* c;
    int Mseq=0;
    while (*(c = a+strcspn(a, ",\\}")) != '\0') {
      switch (*c) {
      case ',':
        *c = '\0';
        Puts(a);
        Puts("</code>, <code>");
        *c = ',';
        a = c+1;
	break;
      case '\\': // TeX command.  We are only interested in \Mseq
	*c='\0';
	Puts(a);
	*c='\\';
	if(strncmp(c,"\\Mseq{",strlen("\\Mseq{"))==0) {
	  Mseq++;
	  a = c+strlen("\\Mseq{");
	} else {
	  if(*(c+1) == '}') {
	    Puts("}");
	    a = c+2;
	  } else {
	    Puts("\\");
	    a = c+1;
	  }
	}
	break;
      case '}':
	*c = '\0';
	Puts(a);
	if(Mseq>0) {
	  Puts("</code>, <code>...");
	  Mseq--;
	}
	*c = '}';
	a = c+1;
	break;
      default:
	fprintf(stderr, "Warning: Strange strcspn result, got %c\n", *c);
	break;
      }
    }
    Puts(a);
    Puts("</code>\n</p>\n");
    gScanner->markPar();
    return true;
}

// MRelatedFunctions lists related functions

static class MRelatedFunctionsCmd : public CmdHdl {
protected:
  const char *title;
public:
  MRelatedFunctionsCmd(const char *s) : CmdHdl(0, 1), title(s) {}
  virtual bool translate(Stream *);
  virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mRelatedFunctionsCmd("relfuncs"), mRelatedDomainsCmd("reldoms"),
  mSuperDomainCmd("supdom");

bool MRelatedFunctionsCmd::translate(Stream *s) {
  localScanner _( &gMupScanner );
  return CmdHdl::translate(s);
}

bool
MRelatedFunctionsCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    // '<h3>#(relfuncs):</h3><p><code>#1</code></p>'
    if (gIe4) {
        Puts(
            "<h4 class=\"sect\"><a onclick=\"expd(relf,aRelf)\" href=\"#Z\">\n"
            "<img src=\"../PICTURES/open.gif\" id=\"aRelf\" border=0>");
        Puts(getDef(title));
        Puts(
            "</a></h4>\n"
            "<div class=\"exp\" id=\"relf\" style=\"display:;\"><p><code>");
    } else {
        Puts(
            "<h4 class=\"sect\"><img src=\"../PICTURES/closed.gif\" border=0> ");
        Puts(getDef(title));
        Puts(
            "</h4><p><code>");
    }
    gScanner->donePar();

    SubStream str(ar[0], "");
    char* a_copy = strdup(ar[0]->cont());
    char* a = a_copy;
    char* c;
    // Das hier kommt immer noch nicht mit \% klar, die sollten aber
    // in MRelatedfunctions eh nicht auftauchen.
    while ((c = strchr(a, '%')) != 0) {
      char *tmp = strchr(c, '\n');
      if(NULL == tmp) {
	*c = '\0';
	break;
      } else {
	// Von % bis Zeilenende rausschneiden
	tmp = strdup(tmp+1);
	strcpy(c, tmp);
	free(tmp);
      }
    }

    while ((c = strchr(a, ',')) != 0) {
        *c = '\0';
        str = a;
        if (*a != 0) putRef((SubStream*)NULL, &str);
        Puts("</code>, <code>");
        *c = ',';
        a = c+1+strspn(c+1," \t\n");
    }
    str = a;
    if (*a != 0) putRef((SubStream*)NULL, &str);
    free(a_copy);
    
    if (gIe4) {
        Puts("</code></p></div>");
    } else {
        Puts("</code></p>");
    }
    return true;
}

// MKeywords lists keywords

static class MKeywordsCmd : public CmdHdl {
public:
    MKeywordsCmd() : CmdHdl(0, 1) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mKeywordsCmd;

bool
MKeywordsCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *ar[]) {
    // '<h3>#(keywords):</h3><p><code>#1</code></p>'
    if (gIe4) {
        Puts(
            "<h4 class=\"sect\"><a onclick=\"expd(keyw,aKeyw)\" href=\"#Z\">\n"
            "<img src=\"../PICTURES/open.gif\" id=\"aKeyw\" border=0>");
        Puts(getDef("keywords"));
        Puts(
            "</a></h4>\n"
            "<div class=\"exp\" id=\"keyw\" style=\"display:;\"><p>");
	gScanner->donePar();
        gScanner->translate(ar[0]);
        Puts("</p></div>");
	gScanner->markPar();
    } else {
        Puts(
            "<h4 class=\"sect\"><img src=\"../PICTURES/closed.gif\" border=0> ");
        Puts(getDef("keywords"));
        Puts(
            "</h4><p>");
	gScanner->donePar();
        gScanner->translate(ar[0]);
        Puts("</p>");
	gScanner->markPar();
    }
    return true;
}

// Mmath and Mdmath require the Mup Scanner
class MmathCmd : public CmdHdl {
protected:
  const char * const prefix;
  const char * const postfix;
public:
    MmathCmd(const char * const pre, const char * const post) :
      CmdHdl(0, 2), prefix(pre), postfix(post) {}
    virtual bool translateArgs(Stream *, SubStream *[], int, SubStream *[]);
} mmathCmd("<var>","</var>"), mdmathCmd("<pre>","</pre>");

bool
MmathCmd::translateArgs(Stream *s, SubStream *oa[], int no, SubStream *a[]) {
    localScanner _(&gMupScanner);
    Puts(prefix);
    gScanner->translate(a[0]);
    Puts(postfix);
    return true;
}


// init the command handlers

void
initMhelpCmds() {
    gCmds.push("Mname", &mnameCmd);
    gCmds.push("Mthis", &mthisCmd);
    gCmds.push("Mkernel", &mkernelCmd);
    gCmds.push("Margs", &margsCmd);
    gCmds.push("Mup", &mupCmd);
    gCmds.push("MupToDo", &mupToDoCmd);
    gCmds.push("Mslot", &mslotCmd);
    gCmds.push("Mex", &mexCmd);
    gCmds.push("Mdom", &mupCmd);
    gCmds.push("Mvar", &mupCmd);
    gCmds.push("Mfunc", &mupCmd);
    gCmds.push("Mupc", &mupcCmd);
    gCmds.push("Mlanggerman", &mlanggermanCmd);
    gCmds.push("Mlangenglish", &mlangenglishCmd);
    gCmds.push("Mhref", &mhrefCmd);
    gCmds.push("Mnoderef", &mhrefCmd); //obselete
    gCmds.push("Mtocline", &mtoclineCmd);
    gCmds.push("Maddtoconts", &maddtocontsCmd);
    gCmds.push("Mcontents", &mcontentsCmd);
    gCmds.push("Mconts", &mcontsCmd);
    gCmds.push("Mversion", &mversionCmd);
    gCmds.push("Moutput", &moutputCmd);
    gCmds.push("Mmenuhref", &mmenuhrefCmd);
    gCmds.push("Mnodepageref", &mmenuhrefCmd); //obselete
    gCmds.push("Mbookbuttons", &mbookbuttonsCmd);
    gCmds.push("Mvvout", &gVerbCmd);
    gCmds.push("Mvvoutn", &gVerbCmd);
    gCmds.push("Minput", &minputCmd);
    gCmds.push("Mincpath", &mincpathCmd);
    gCmds.push("Mowninput", &minputCmd); //obselete
    gCmds.push("Mentryref", &mentryrefCmd);
    gCmds.push("Mpacktitle", &mpacktitleCmd);
    // New for 1.5
    gCmds.push("de", &deCmd);
    gCmds.push("eng", &engCmd);
    gCmds.push("call", &mcallCmd);
    gCmds.push("Mcall", &mcallCmd);
    gCmds.push("MCall", &mcallCmd); // actually, it's slightly different.
    gCmds.push("MDeclareOption", &mDeclareOptionHdl);
    gCmds.push("MOverloading", &mOverloadingCmd);
    gCmds.push("MRelatedFunctions", &mRelatedFunctionsCmd);
    gCmds.push("MRelatedDomains", &mRelatedDomainsCmd);
    gCmds.push("MSuperDomain", &mSuperDomainCmd);
    gCmds.push("MKeywords", &mKeywordsCmd);
    gCmds.push("Mmath", &mmathCmd);
    gCmds.push("Mdmath", &mdmathCmd);
}

// end of file
