/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */


// Stack.h -- stacks and stack links

#ifndef _stack_h_
#define _stack_h_

#include "type.h"
#include "error.h"

class StackLink {
    friend class StackBase;
    StackLink *nextPtr;
    virtual void push() {};
    virtual void pop() {};
public:
    StackLink() { nextPtr = NULL; }
};

class StackBase {
    StackLink *topPtr;
public:
    StackBase() { topPtr = NULL; }
    StackBase(StackLink *e) {
        topPtr = e;
    }
    void push(StackLink *e) {
        e->nextPtr = topPtr;
        e->push();
        topPtr = e;
    }
    StackLink *pop() {
        StackLink *e = topPtr;
        if (e == NULL) fatal("internal (Stack::pop)");
        e->pop();
        topPtr = topPtr->nextPtr;
        return e;
    }
    StackLink* top() { return topPtr; }
    bool empty() { return topPtr == NULL; }
};

template<class T>
class Stack : public StackBase {
public:
    Stack() : StackBase() {}
    Stack(T *e) : StackBase((StackLink *) e) {}
    void push(T *e) { StackBase::push((StackLink *) e); }
    T *pop() { return (T *) StackBase::pop(); }
    T* top() { return (T *) StackBase::top(); }
};

#endif // _stack_h_

