/*
 * This file is Copyright 2002 by SciFace Software GmbH.
 * It is part of a public prerelease, but may *not* be redistributed.
 * See the accompanying README file for details.
 */


// Stream.h -- character streams

#ifndef _stream_h_
#define _stream_h_

#include <stdio.h>
#include <string>
#include "type.h"

const char kEndChar = '\0';

class StreamPos {
public:
    union _pos { long fpos; char *cpos; } pos;
    long lineNo;
};

class Stream {
protected:
    long lineNo;
    char *fnam;
public:
    Stream();
    virtual ~Stream() = 0;
    virtual char get() = 0;
    virtual StreamPos pos() = 0;
    virtual bool getRange(StreamPos, char *, long) = 0;
    virtual bool getRange(StreamPos, std::string&) = 0;
    virtual void unget(char c) = 0;
    virtual bool end() = 0;
    virtual void reset() = 0;
    virtual void fatal(char *, char * = "", char * = "");
    virtual void error(char *, char * = "", char * = "");
    virtual void warning(char *, char * = "", char * = "");
    virtual long line() { return lineNo; }
    virtual char *name() { return fnam; }
};

class FileStream : public Stream {
protected:
    FILE *inFile;
    long startPos, actPos, actLine;
public:
    FileStream(FILE *, char *);
    ~FileStream();
    char get();
    StreamPos pos();
    bool getRange(StreamPos, char *, long);
    bool getRange(StreamPos, std::string&);
    void unget(char c);
    bool end();
    void reset();
    void save();
    void restore();
};

class SubStream : public Stream {
    char *contPtr, *actPos;
    long startLine;
public:
    SubStream();
    SubStream(Stream *, char *);
    SubStream(Stream *, char *, StreamPos *p);
    SubStream(Stream *, std::string&);
    SubStream(Stream *, std::string&, StreamPos *p);
    ~SubStream() { delete contPtr; }
    SubStream& operator=(SubStream *);
    SubStream& operator=(const char *);
    char get();
    StreamPos pos();
    bool getRange(StreamPos, char *, long);
    bool getRange(StreamPos, std::string&);
    bool copy(char *, int);
    char *cont() { return contPtr; }
    void unget(char c);
    bool end();
    void reset();
    bool empty() { return (*contPtr == '\0'); }
};

extern SubStream gNilStream;

#endif // _stream_h_

